##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from pathlib import Path
from tempfile import TemporaryDirectory

from f4ratk.directories import Cache


def given_path_when_register_should_create_all_directories_in_path():
    with TemporaryDirectory(prefix='f4ratk_test_tmp_') as parent:
        target_path = Path(parent, 'intermediate', 'target')

        Cache.register(str(target_path))

        assert target_path.exists()
        assert target_path.is_dir()


def given_cache_when_file_with_name_requested_should_append_name_to_cache_path():
    with TemporaryDirectory(prefix='f4ratk_test_tmp_') as parent:
        result = Cache.register(parent).file('target')
        assert result == Path(parent, 'target')



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from unittest.mock import Mock

from numpy import dtype
from pandas import Period, PeriodDtype, PeriodIndex
from pandas.tseries.offsets import BusinessDay, MonthEnd
from pandas_datareader.famafrench import FamaFrenchReader
from pytest import mark, raises
from pytest_mock import MockerFixture

from f4ratk.domain import Frequency, Region
from f4ratk.fama import FamaReader
from f4ratk.shared import Normalizer
from tests.conftest import FamaFactory


class TestFamaReader:
    @mark.parametrize(
        'region',
        (region for region in Region),
    )
    def given_any_valid_region_and_monthly_frequency_should_return_data_with_monthly_period_index(  # noqa: E501
        self,
        region: Region,
    ):
        returns = FamaReader(None).fama_data(region=region, frequency=Frequency.MONTHLY)

        assert isinstance(returns.index, PeriodIndex)
        assert returns.index.dtype == PeriodDtype(freq=MonthEnd())

    @mark.parametrize(
        'region',
        (region for region in Region if region != Region.EMERGING),
    )
    def given_any_valid_region_and_daily_frequency_should_return_data_with_daily_period_index(  # noqa: E501
        self,
        region: Region,
    ):
        returns = FamaReader(Normalizer()).fama_data(
            region=region, frequency=Frequency.DAILY
        )

        assert isinstance(returns.index, PeriodIndex)
        assert returns.index.dtype == PeriodDtype(freq=BusinessDay())

    def given_emerging_region_for_daily_frequency_should_raise_value_error(self):
        with raises(ValueError):
            FamaReader(None).fama_data(
                region=Region.EMERGING, frequency=Frequency.DAILY
            )

    def given_any_region_for_any_frequency_should_return_five_factors_plus_momentum(
        self,
    ):
        returns = FamaReader(None).fama_data(
            region=Region.US, frequency=Frequency.MONTHLY
        )

        assert list(returns.columns) == ['MKT', 'SMB', 'HML', 'RMW', 'CMA', 'RF', 'WML']

    def given_data_missing_for_dates_when_reading_data_should_remove_missing_data(
        self, create_fama: FamaFactory, mocker: MockerFixture
    ):

        returns = create_fama(
            ('2020-04', -99.99),
            ('2020-05', -90.00),
            ('2020-06', -99.99),
        )

        reader_factory = mocker.patch('f4ratk.fama.fama_french_reader')

        fama_reader = Mock(spec_set=FamaFrenchReader)
        fama_reader.read.return_value = {0: returns}
        reader_factory.return_value = fama_reader

        result = FamaReader(normalizer=Normalizer()).fama_data(
            region=Region.EMERGING, frequency=Frequency.MONTHLY
        )

        assert len(result) == 1
        assert all([t is dtype(float) for t in result.dtypes])
        assert result.index[0] == Period('2020-05', 'M')



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from f4ratk.analyze.api import DataAnalyzer
from f4ratk.analyze.evaluation import EvaluatedResults
from f4ratk.domain import AnalysisConfig
from f4ratk.fama import FamaReader
from f4ratk.history import History
from f4ratk.ticker.api import Stock, TickerAnalyzer, TickerReader


class TickerAnalyzerAdapter(TickerAnalyzer):
    def __init__(
        self,
        fama_reader: FamaReader,
        analyzer: DataAnalyzer,
        history: History,
        ticker_reader: TickerReader,
    ):
        self._fama_reader = fama_reader
        self._analyzer = analyzer
        self._history = history
        self._ticker_reader = ticker_reader

    def analyze_ticker_symbol(
        self, stock: Stock, analysis_config: AnalysisConfig
    ) -> EvaluatedResults:
        data = self._ticker_reader.read_ticker(stock=stock, frame=analysis_config.frame)

        historic_returns = self._history.annualized_returns(
            region=analysis_config.region
        )

        return self._analyzer.analyze(
            data,
            self._fama_reader.fama_data(
                region=analysis_config.region, frequency=analysis_config.frame.frequency
            ),
            historic_returns,
        )



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from os import getenv
from pathlib import Path
from typing import Union

from flask_caching import Cache

from f4ratk.directories import cache


def locate_cache_dir() -> Union[str, Path]:
    return getenv('CACHE_DIR', cache.file('server'))


def create_cache() -> Cache:
    cache_type = getenv('CACHE_TYPE')
    cache_dir = locate_cache_dir() if cache_type == 'FileSystemCache' else None

    config = dict()

    if cache_type:
        config['CACHE_TYPE'] = cache_type

    if cache_dir:
        config['CACHE_DIR'] = cache_dir

    return Cache(config=config)


server_cache = create_cache()



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from f4ratk.analyze.api import DataAnalyzer
from f4ratk.analyze.evaluation import EvaluatedResults
from f4ratk.domain import AnalysisConfig
from f4ratk.fama import FamaReader
from f4ratk.file.api import FileAnalyzer, FileReader
from f4ratk.file.reader import FileConfig
from f4ratk.history import History


class FileAnalyzerAdapter(FileAnalyzer):
    def __init__(
        self,
        fama_reader: FamaReader,
        analyzer: DataAnalyzer,
        history: History,
        file_reader: FileReader,
    ):
        self._fama_reader = fama_reader
        self._analyzer = analyzer
        self._history = history
        self._file_reader = file_reader

    def analyze_file(
        self, file_config: FileConfig, analysis_config: AnalysisConfig
    ) -> EvaluatedResults:
        data = self._file_reader.read(
            file_config=file_config,
            frame=analysis_config.frame,
        )

        historic_returns = self._history.annualized_returns(
            region=analysis_config.region
        )

        return self._analyzer.analyze(
            data,
            self._fama_reader.fama_data(
                region=analysis_config.region, frequency=analysis_config.frame.frequency
            ),
            historic_returns=historic_returns,
        )



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from click import group
from flask import Flask, jsonify
from flask.cli import FlaskGroup

# noinspection PyPackageRequirements
from werkzeug.exceptions import HTTPException, NotFound

from f4ratk.infrastructure import configure_logging, instantiate_dependencies
from f4ratk.ticker.api import NoTickerData
from f4ratk.web.infrastructure import server_cache


def create_app() -> Flask:
    configure_logging(verbose=False, server=True)
    instantiate_dependencies()

    app = Flask('f4ratk-server', static_folder=None)

    server_cache.init_app(app)

    @app.errorhandler(HTTPException)
    def error_500(exception: HTTPException):
        return (
            jsonify({"error": exception.description}),
            exception.code,
            {'Content-Type': 'application/json'},
        )

    @app.errorhandler(NoTickerData)
    def error_symbol_unknown(exception: NoTickerData):
        return (
            jsonify({"error": str(exception)}),
            NotFound.code,
            {'Content-Type': 'application/json'},
        )

    with app.app_context():
        # noinspection PyUnresolvedReferences
        from f4ratk.web import controller  # noqa: F401

    return app


@group(cls=FlaskGroup, create_app=create_app)
def server_main():
    pass



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from pytest import mark

from f4ratk.domain import Frequency, Region
from f4ratk.fama import ReturnsRequest


@mark.parametrize(
    'region, expected_source_ff, expected_source_mom',
    (
        (
            Region.DEVELOPED,
            'Developed_5_Factors_Daily',
            'Developed_Mom_Factor_Daily',
        ),
        (
            Region.DEVELOPED_EX_US,
            'Developed_ex_US_5_Factors_Daily',
            'Developed_ex_US_Mom_Factor_Daily',
        ),
        (
            Region.US,
            'F-F_Research_Data_5_Factors_2x3_daily',
            'F-F_Momentum_Factor_daily',
        ),
        (
            Region.EU,
            'Europe_5_Factors_Daily',
            'Europe_Mom_Factor_Daily',
        ),
    ),
)
def given_any_valid_region_and_daily_frequency_when_creating_request_should_map_to_remote_fama_data_set_names(  # noqa: E501
    region: Region,
    expected_source_ff: str,
    expected_source_mom: str,
):
    request = ReturnsRequest.of(region=region, frequency=Frequency.DAILY)

    assert request.frequency == Frequency.DAILY
    assert request.five.value == expected_source_ff
    assert request.momentum.value == expected_source_mom


@mark.parametrize(
    'region, expected_source_ff, expected_source_mom',
    (
        (
            Region.DEVELOPED,
            'Developed_5_Factors',
            'Developed_Mom_Factor',
        ),
        (
            Region.DEVELOPED_EX_US,
            'Developed_ex_US_5_Factors',
            'Developed_ex_US_Mom_Factor',
        ),
        (
            Region.US,
            'F-F_Research_Data_5_Factors_2x3',
            'F-F_Momentum_Factor',
        ),
        (
            Region.EU,
            'Europe_5_Factors',
            'Europe_Mom_Factor',
        ),
        (
            Region.EMERGING,
            'Emerging_5_Factors',
            'Emerging_MOM_Factor',
        ),
    ),
)
def given_any_valid_region_and_monthly_frequency_when_creating_request_should_map_to_remote_fama_data_set_names(  # noqa: E501
    region: Region,
    expected_source_ff: str,
    expected_source_mom: str,
):
    request = ReturnsRequest.of(region=region, frequency=Frequency.MONTHLY)

    assert request.frequency == Frequency.MONTHLY
    assert request.five.value == expected_source_ff
    assert request.momentum.value == expected_source_mom



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from datetime import date

from pandas import PeriodDtype, PeriodIndex
from pandas.tseries import offsets

from f4ratk.domain import Currency, Frequency
from f4ratk.exchange import ExchangeReader


def given_daily_frequency_when_reading_data_should_return_daily_exchange_rate():
    result = ExchangeReader(frequency=Frequency.DAILY).exchange_data(
        currency=Currency.EUR, start=date(2020, 9, 16), end=date(2020, 9, 16)
    )

    assert isinstance(result.index, PeriodIndex)
    assert result.index.dtype == PeriodDtype(freq=offsets.BusinessDay())

    assert result['Exchange Rate']['2020-09-16'] == 1.1835


def given_monthly_frequency_when_reading_data_should_return_monthly_exchange_rate():
    result = ExchangeReader(frequency=Frequency.MONTHLY).exchange_data(
        currency=Currency.EUR, start=date(2020, 9, 1), end=date(2020, 9, 30)
    )

    assert isinstance(result.index, PeriodIndex)
    assert result.index.dtype == PeriodDtype(freq=offsets.MonthEnd())

    assert len(result) == 1
    assert result['Exchange Rate']['2020-09'] == 1.1785


def should_rename_exchange_rate_column_to_generic_name():
    result = ExchangeReader(frequency=Frequency.DAILY).exchange_data(
        currency=Currency.EUR, start=date(2020, 9, 16), end=date(2020, 9, 16)
    )

    assert list(result.columns) == ['Exchange Rate']
